# -*- coding: utf-8 -*-


class MoconException(Exception):
    pass


class InvalidModelException(MoconException):
    pass


class InvalidKeyException(MoconException):
    pass


class PrimaryKeyNotFound(MoconException):
    pass


class PrimaryKeyDuplicated(MoconException):
    pass


class NoSourceError(MoconException):
    pass


class NotSupportedSource(MoconException):
    pass



# -*- coding: utf-8 -*-
from typing import Any, Optional, Union
from urllib.parse import SplitResult, parse_qsl, unquote, urlsplit

from pymongo import MongoClient
from redis import Redis

from kvom.exceptions import NotSupportedSource


class Backend:
    def get(self, key: str) -> Union[str, None]:
        raise NotImplementedError

    def set(self, key: str, value: str) -> bool:
        raise NotImplementedError

    def delete(self, key: str) -> bool:
        raise NotImplementedError


class RedisBackend(Backend):
    def __init__(self, url: Union[str, "SourceURL"], **options: Any) -> None:
        self._url = SourceURL(url)
        self._options = options

    def connection(self) -> "Redis":
        client = Redis(
            host=self._url.hostname,
            port=self._url.port,
            db=int(self._url.database),
            password=self._url.password,
            decode_responses=True,
            **self._options,
        )
        return client

    def get(self, key: str) -> Union[str, None]:
        return self.connection().get(key)

    def set(self, key: str, value: str) -> bool:
        return self.connection().set(key, value)

    def delete(self, key: str) -> bool:
        return bool(self.connection().delete(key))


class MongoBackend(Backend):
    def __init__(
        self, url: Union[str, "SourceURL"], document: str = "kvom", **options: Any
    ) -> None:
        self._url = SourceURL(url)
        self._options = options
        self._document_name = document

    def connection(self) -> "MongoClient":
        client = MongoClient(
            host=self._url.hostname,
            port=self._url.port,
            **self._options,
        )
        return client[self._url.database]

    @property
    def _document(self) -> "MongoClient":
        return self.connection()[self._document_name]

    def get(self, key: str) -> Union[str, None]:
        return self._document.find_one({"key": key})["value"]

    def set(self, key: str, value: str) -> bool:
        return self._document.insert_one({"key": key, "value": value})

    def delete(self, key: str) -> bool:
        return self._document.delete_one({"key": key})


class Source:
    SUPPORTED_SOURCE = {
        "redis": RedisBackend,
        "mongodb": MongoBackend,
    }

    def __init__(self, url: Union[str, "SourceURL"], **options):
        self.url = SourceURL(url)
        self.options = options
        self._backend = self.switch_source()

    def switch_source(self):
        Client = self.SUPPORTED_SOURCE.get(self.url.dialect)
        if not Client:
            raise NotSupportedSource(f"{self.url.dialect} is not supported")
        return Client(self.url, **self.options)

    def get(self, key: str) -> str:
        return self._backend.get(key)

    def set(self, key: str, value: str) -> bool:
        return self._backend.set(key, value)

    def delete(self, key: str) -> bool:
        return self._backend.delete(key)


def _EmptyNetloc():
    def __bool__(self) -> bool:
        return True


class SourceURL:
    def __init__(self, url: Union[str, "SourceURL"]):
        if isinstance(url, SourceURL):
            self._url: str = url._url
        elif isinstance(url, str):
            self._url = url
        else:
            raise TypeError(
                f"Invalid type for DatabaseURL. "
                f"Expected str or DatabaseURL, got {type(url)}"
            )

    @property
    def components(self) -> SplitResult:
        if not hasattr(self, "_components"):
            self._components = urlsplit(self._url)
        return self._components

    @property
    def scheme(self) -> str:
        return self.components.scheme

    @property
    def dialect(self) -> str:
        return self.components.scheme.split("+")[0]

    @property
    def driver(self) -> str:
        if "+" not in self.components.scheme:
            return ""
        return self.components.scheme.split("+", 1)[1]

    @property
    def userinfo(self) -> Optional[bytes]:
        if self.components.username:
            info = self.components.username
            if self.components.password:
                info += ":" + self.components.password
            return info.encode("utf-8")
        return None

    @property
    def username(self) -> Optional[str]:
        if self.components.username is None:
            return None
        return unquote(self.components.username)

    @property
    def password(self) -> Optional[str]:
        if self.components.password is None:
            return None
        return unquote(self.components.password)

    @property
    def hostname(self) -> Optional[str]:
        return (
            self.components.hostname
            or self.options.get("host")
            or self.options.get("unix_sock")
        )

    @property
    def port(self) -> Optional[int]:
        return self.components.port

    @property
    def netloc(self) -> Optional[str]:
        return self.components.netloc

    @property
    def database(self) -> str:
        path = self.components.path
        if path.startswith("/"):
            path = path[1:]
        return unquote(path)

    @property
    def options(self) -> dict:
        if not hasattr(self, "_options"):
            self._options = dict(parse_qsl(self.components.query))
        return self._options

    def replace(self, **kwargs: Any) -> "SourceURL":
        if (
            "username" in kwargs
            or "password" in kwargs
            or "hostname" in kwargs
            or "port" in kwargs
        ):
            hostname = kwargs.pop("hostname", self.hostname)
            port = kwargs.pop("port", self.port)
            username = kwargs.pop("username", self.components.username)
            password = kwargs.pop("password", self.components.password)

            netloc = hostname
            if port is not None:
                netloc += f":{port}"
            if username is not None:
                userpass = username
                if password is not None:
                    userpass += f":{password}"
                netloc = f"{userpass}@{netloc}"

            kwargs["netloc"] = netloc

        if "database" in kwargs:
            kwargs["path"] = "/" + kwargs.pop("database")

        if "dialect" in kwargs or "driver" in kwargs:
            dialect = kwargs.pop("dialect", self.dialect)
            driver = kwargs.pop("driver", self.driver)
            kwargs["scheme"] = f"{dialect}+{driver}" if driver else dialect

        if not kwargs.get("netloc", self.netloc):
            # Using an empty string that evaluates as True means we end up
            # with URLs like `sqlite:///database` instead of `sqlite:/database`
            kwargs["netloc"] = _EmptyNetloc()

        components = self.components._replace(**kwargs)
        return self.__class__(components.geturl())

    @property
    def obscure_password(self) -> str:
        if self.password:
            return self.replace(password="********")._url
        return self._url

    def __str__(self) -> str:
        return self._url

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({repr(self.obscure_password)})"

    def __eq__(self, other: Any) -> bool:
        return str(self) == str(other)



# -*- coding: utf-8 -*-
from typing import AbstractSet, Any, Dict, Mapping, Optional, Union

from pydantic.fields import FieldInfo as PydanticFieldInfo
from pydantic.fields import NoArgAnyCallable, Undefined


class FieldInfo(PydanticFieldInfo):
    def __init__(self, default: Any = Undefined, **kwargs: Any) -> None:
        primary_key = kwargs.pop("primary_key", False)
        super().__init__(default=default, **kwargs)
        self.primary_key = primary_key


def Field(
    default: Any = Undefined,
    *,
    default_factory: Optional[NoArgAnyCallable] = None,
    alias: str = None,
    title: str = None,
    description: str = None,
    exclude: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    include: Union[
        AbstractSet[Union[int, str]], Mapping[Union[int, str], Any], Any
    ] = None,
    const: bool = None,
    gt: float = None,
    ge: float = None,
    lt: float = None,
    le: float = None,
    multiple_of: float = None,
    min_items: int = None,
    max_items: int = None,
    min_length: int = None,
    max_length: int = None,
    allow_mutation: bool = True,
    regex: str = None,
    primary_key: bool = False,
    schema_extra: Optional[Dict[str, Any]] = None,
) -> Any:
    current_schema_extra = schema_extra or {}
    field_info = FieldInfo(
        default,
        default_factory=default_factory,
        alias=alias,
        title=title,
        description=description,
        exclude=exclude,
        include=include,
        const=const,
        gt=gt,
        ge=ge,
        lt=lt,
        le=le,
        multiple_of=multiple_of,
        min_items=min_items,
        max_items=max_items,
        min_length=min_length,
        max_length=max_length,
        allow_mutation=allow_mutation,
        regex=regex,
        primary_key=primary_key,
        **current_schema_extra,
    )
    field_info._validate()
    return field_info



# -*- coding: utf-8 -*-
import secrets
from typing import Any, ClassVar, Optional, TypeVar

from pydantic import BaseModel as PydanticBaseModel
from pydantic import validator
from pydantic.main import ModelMetaclass

from kvom.exceptions import NoSourceError, PrimaryKeyDuplicated, PrimaryKeyNotFound
from kvom.field import Field, FieldInfo
from kvom.source import Source
from kvom.field import Field

__all__ = ["BaseModel"]

T = TypeVar("T", bound="BaseModel")


class BaseMeta:
    # data source
    source: Optional[Source] = None

    # the model encoding format when saving to the store
    encoding: str = "utf-8"

    prefix: Optional[str] = None

    primary_key: Optional[str] = None

    # is support embedded model
    embedded: bool = False


def _set_meta_default(cls, meta, base_meta):
    if not getattr(meta, "encoding", None):
        meta.encoding = getattr(base_meta, "encoding", "utf-8")

    if not getattr(meta, "prefix", None):
        meta.prefix = f"{cls.__module__}:{cls.__name__}".lower()

    if not getattr(meta, "embedded", None):
        meta.embedded = getattr(base_meta, "embedded", False)

    if not getattr(meta, "source", None):
        meta.source = getattr(base_meta, "source", None)


class BaseModelMeta(ModelMetaclass):
    _meta: BaseMeta

    def __new__(mcs, name, bases, attrs, **kwargs):
        cls = super().__new__(mcs, name, bases, attrs, **kwargs)
        meta = attrs.get("Meta", None)

        # if cls not defined Meta, get from parent
        meta = meta or getattr(cls, "Meta", None)
        base_meta = getattr(cls, "_meta", None)
        # no inherited, defined Meta
        if meta and meta != BaseMeta and meta != base_meta:
            cls.Meta = meta
            cls._meta = meta
        # inherited Meta
        elif base_meta:
            cls._meta = type(
                f"{cls.__name__}Meta", (base_meta,), dict(base_meta.__dict__)
            )
            cls.Meta = cls._meta
        # no defined Meta, no inherited, use default
        else:
            cls._meta = type(
                f"{cls.__name__}Meta", (BaseMeta,), dict(BaseMeta.__dict__)
            )
            cls.Meta = cls._meta

        for f_name, field in cls.__fields__.items():
            if isinstance(field.field_info, FieldInfo):
                cls._meta.primary_key = f_name

        # set Meta default value if there is no defined
        _set_meta_default(cls, cls._meta, base_meta)

        return cls


class BaseModel(PydanticBaseModel, metaclass=BaseModelMeta):
    """
    Example:

    class UserModel(BaseModel):
        class Meta:
            can_edit = False

        name: str
        age: int

    user = UserModel(name="John", age=18)
    user.save()

    """
    pk: Optional[str] = Field(default=None, primary_key=True)

    pk: Optional[str] = Field(default=None, primary_key=True)

    Meta = BaseMeta
    identity: ClassVar[str]

    def __init__(__pydantic_self__, **data: Any) -> None:
        super().__init__(**data)
        __pydantic_self__.validate_source()
        __pydantic_self__.validate_pk()

    @property
    def key(self):
        prefix = getattr(self._meta, "prefix", "")
        primary_key = getattr(self._meta, "primary_key")
        return f"{prefix}:{primary_key}".strip(":")

    @classmethod
    def validate_source(cls):
        if cls._meta.source is None or not isinstance(cls._meta.source, Source):
            raise NoSourceError("Model must have a Source client")

    @classmethod
    def validate_pk(cls):
        primary_keys = 0
        for name, field in cls.__fields__.items():
            if getattr(field.field_info, "primary_key", None):
                primary_keys += 1
        if primary_keys == 0:
            raise PrimaryKeyNotFound("You must define a primary key for the model")
        elif primary_keys > 1:
            raise PrimaryKeyDuplicated(
                "You must define only one primary key for a model"
            )

    @validator("pk", always=True, allow_reuse=True)
    def pk_or_default(cls, v):
        if not v:
            v = secrets.token_hex(4)
        return v

    @classmethod
    def get(cls, key: str) -> Optional["BaseModel"]:
        data = cls._meta.source.get(key)
        if not data:
            return None
        return cls.parse_raw(data, encoding=cls._meta.encoding)

    def save(self) -> bool:
        cls = self.__class__
        return cls._meta.source.set(self.key, self.json())

    def delete(self) -> bool:
        cls = self.__class__
        return cls._meta.source.delete(self.key)
